
=head1 NAME

ll4model.pm

=head1 SYNOPSIS

This module can be used as a foundation for calculating all the 
Log Likelihood scores for modeling a 4-dimensional ngram. This
model is used by model4.pm. In particular this module provides 
functions that give convenient access to  4-d (i.e., 4gram) 
frequency counts as created by count.pl, as well as  some degree 
of error handling that verifies the data.  

To be used in a measure module that is to be used by statistic.pl, the  
functions provided in this module must be embedded within other functions  
that adhere to the standards and naming convention described in  
Docs/NewStats.txt and is briefly summarized below.

=head1 DESCRIPTION

The functions in this module retrieve observed 4gram frequency counts and  
marginal totals, compute expected values and the log likelihhood scores
for each of the possible assumptions. They also provide support for error 
checking of the output produced by count.pl. 

The Log Likelhood measure is the ratio between how often an ngram actually 
occurred compared to how often it would be expected occur. In this measure, 
you are comparing the observed and expected values for each "cell" in a 
contingency table that represents the counts of your observed ngram. 

In the 2 dimensional case, there is only one possible model that 
you can compare to, and that is the model of independence. The log 
likelihood calculates the observed count of your bigram, and compares 
it to the counts that would be expected if the words in the bigram were 
statistically independent. Statistical independence is the probability 
that two words occur together is equal to the product of the their 
individual probabilites. More formally, this can be defined as 
P(word1, word2) = P(word1) * P(word2)

The score that log-likelihood measure produces reflects the degree to 
which the observed and expected values diverge. A log likelihood score 
of zero implies that the data fits perfectly into the independence model 
and the expected values are equal to the observed values. Therefore, the 
higher the log likelihood score, the less the tokens in the ngram 
appear to be independent from each other. 

Calculating the expected values based on independence can be carried to 
4grams but as the dimensions of your contingency table grow, so does the 
number of models in which the words can be compared to. The expected values 
for a 4grams can be based on 13 other models listed below:
 
Model 1 : P(w1 w2 w3 w4) = ( P(w1) P(w2 w3 w4) )          
Model 2 : P(w1 w2 w3 w4) = ( P(w1 w2) P(w3) P(w4) )       
Model 3 : P(w1 w2 w3 w4 ) = ( P(w1) P(w2) P(w3 w4) )
Model 4 : P(w1 w2 w3 w4) = ( P(w1) P(w2 w3) P(w4) )        
Model 5 : P(w1 w2 w3 w4) = ( P(w1 w2 w3) P(w4) )          
Model 6 : P(w1 w2 w3 w4) = ( P(w1) P(w2 w4) P(w3) )       
Model 7 : P(w1 w2 w3 w4) = ( P(w1 w3) P(w2 w4) )
Model 8 : P(w1 w2 w3 w4) = ( P(w1 w4) P(w2 w3) )
Model 9 : P(w1 w2 w3 w4) = ( P(w1 w3 w4) P(w2)
Model 10: P(w1 w2 w3 w4) = ( P(w1 w2 w4) P(w3) )
Model 11: P(w1 w2 w3 w4) = ( P(w1 w2) P(w3 w4) )
Model 12: P(w1 w2 w3 w4) = ( P(w1 w4) P(w2) P(w3) )
Model 13: P(w1 w2 w3 w4) = ( P(w1 w3) P(w2) P(w4)         
Model 14: P(w1 w2 w3 w4 ) = ( P(w1) P(w2) P(w3) P(w4) ) 

Each of these models will result in different expected values and 
log likelihood scores. With 4gram or 4d measures we use a 4x4 contingency 
table to store the  frequency counts associated with each word in the 4gram, 
as well as the  number of times the 4gram occurs. The notation we employ is 
as follows:

Marginal Frequencies:
    
 n1ppp = the number of ngrams where the first word is word1.
 np1pp = the number of ngrams where the second word is word2.
 npp1p = the number of ngrams where the third word is word3
 nppp1 = the number of ngrams where the fourth word is word4
 n2ppp = the number of ngrams where the first word is not word1. 
 np2pp = the number of ngrams where the second word is not word2.
 npp2p = the number of ngrams where the third word is not word3.
 nppp2 = the number of ngrams where the fourth words is not word4

Observed Frequencies:

 n1111 = number of times word1, word2 and word3 occur together in
         their respective positions, joint frequency.
 n1112 = number of times word1, word 2 and word3 occur in their respective
         positions but word4 does not.
 n1121 = number of times word1, word2 and word4 occur in their respective
         positions but word3 does not.
 n1122 = number of times word1 and word2 occur in their repsective positions
         but word3 and word4 do not.
 n1211 = number of times word1, word3 and word4 occur in their respective 
         positions but word2 does not.
 n1212 = number of times word1 and word3 occur in their respective positions
         but word2 and word4 do not.
 n1221 = number of times word1 and word4 occur in their respective positions
         but word2 and word3 do not
 n1222 = number of times word1 occurs in its respective position but word2, 
         word3 and word4 do not.
 n2111 = number of times word2, word3 and word4 occur in their respective 
         positions but word1 does not.
 n2112 = number of times word2 and word3 occur in their respective positions
         but word1 and word4 do not.
 n2121 = number of times word2 and word4 occur in their respective positions
         but word1 and word3 do not.
 n2122 = number of times word2 occurs in its respective position but word1, 
         word3 and word4 do not.
 n2211 = number of times word3 and word4 occur in their respective positions
         but word1 and word2 do not.
 n2212 = number of times word3 occurs in its respective position but word1, 
         word2 and word4 do not.
 n2221 = number of times word4 occurs in its respective position but word1,
         word2, and word3 do not.
 n2222 = number of times neither word1, word2, word3 or word4 occur in their
         respective positions.

Expected Frequencies (Examples for the Independence Assumption):

 m1111 = expected number of times word1, word2 and word3 occur together in
         their respective positions, joint frequency.
 m1112 = expected number of times word1, word 2 and word3 occur in their respective
         positions but word4 does not.
 m1121 = expected number of times word1, word2 and word4 occur in their respective
         positions but word3 does not.
 m1122 = expected number of times word1 and word2 occur in their repsective positions
         but word3 and word4 do not.
 m1211 = expected number of times word1, word3 and word4 occur in their respective 
         positions but word2 does not.
 m1212 = expected number of times word1 and word3 occur in their respective positions
         but word2 and word4 do not.
 m1221 = expected number of times word1 and word4 occur in their respective positions
         but word2 and word3 do not
 m1222 = expected number of times word1 occurs in its respective position but word2, 
         word3 and word4 do not.
 m2111 = expected number of times word2, word3 and word4 occur in their respective 
         positions but word1 does not.
 m2112 = expected number of times word2 and word3 occur in their respective positions
         but word1 and word4 do not.
 m2121 = expected number of times word2 and word4 occur in their respective positions
         but word1 and word3 do not.
 m2122 = expected number of times word2 occurs in its respective position but word1, 
         word3 and word4 do not.
 m2211 = expected number of times word3 and word4 occur in their respective positions
         but word1 and word2 do not.
 m2212 = expected number of times word3 occurs in its respective position but word1, 
         word2 and word4 do not.
 m2221 = expected number of times word4 occurs in its respective position but word1,
         word2, and word3 do not.
 m2222 = expected number of times neither word1, word2, word3 or word4 occur in their
         respective positions.


   
=head2 Functions Included in ll4model.pm

=over 4

=item 1 

The intializeStatistic function performs initialization before 
the actual computation of measures of association begin. It also 
verifies that the input consists of 4grams, and that it has all 
the frequency combinations needed to do the computations. The
frequency combinations that are required for this module are 
n1111,n1ppp,np1pp,npp1p,nppp1,nn11pp,n1p1p,n1pp1,np11p,np1p1,
npp11,n111p,n11p1,n1p11,np111.

initializeStatistic() is passed the following input parameters:

     1) The ngram size. For 4d (4gram) measures this will be 4. 
     2) The total number of 4grams in the corpus (npppp).
     3) The number of frequency combinations. 
     4) A 4-d array containing the frequency combinations.

Each row  of the array in 4) represents a single frequency combination.  
On a given row, the first element denotes the number of indices on this  
row, say 'n'. This is followed by the 'n' values that correspond to the 
indices included in the frequency combination. (For more details on   
frequency combinations, see README.pod). To use this module, the joint
frequency, n11,  as well as the marginal frequencies, n1p and np1, are
required in order to calculate the expected values.

This function does not return any values. If an error occurs, it can be 
detected by statistic.pl using the errorCode and errorString functions 
described below.   

=item 2

The checkObservedValues function takes as input an array containing the 
frequency values for a 4gram as found by count.pl. This will include: 
n1111,n1ppp,np1pp,npp1p,nppp1,nn11pp,n1p1p,n1pp1,np11p,np1p1,npp11,n111p,
n11p1,n1p11,np111. The size of this array is guaranteed to be exactly the 
same as the third parameter passed to the function initializeStatistic() 
function above. 

The checkObservedValues function verifies that the marginal frequencies 
n1ppp,np1pp,npp1p,nppp1,nn11pp,n1p1p,n1pp1,np11p,np1p1,npp11,n111p,
n11p1,n1p11, and np111 are consistent with the value of the joint 
frequency of the 4gram, n1111. If they are not consistent, it sets an 
error code and error message, and the function returns.

If the marginals are valid, it computes the observed values for the   
remaining cells in the 4d table ( n1112, n1121, n1122, n1211, n1212,
n1221, n1222, n2111, n2112, n2121, n2122, n2211, n2212, n2221, n2222 )
based on these marginal totals and the joint frequency and returns 1.

=item 3

The function checkExpectedValues() calculates the expected 
values of the cells in the contingency table based on the marginal 
frequencies and the total sample size for each of assumption
modles as well as the loglikelihood value for each set of the
expected values. The function returns a 1 if everything is okay.

=item 4

The function getMarginalValues will return the marginal frequencies in the
order of: n1ppp,np1pp,npp1p,nppp1,nn11pp,n1p1p,n1pp1,np11p,np1p1,npp11,n111p,
n11p1,n1p11, and np111.

=item 5

The function getTotal4grams() returns the total number of 4grams in the 
corpus (npppp).

= item 6

The function getLogLikelihoodValues() returns the log likelihood scores
for each of the assumption models in the following order: $LogLikelihood1,  
$LogLikelihood2, $LogLikelihood3, $LogLikelihood4, $LogLikelihood5, 
$LogLikelihood6, $LogLikelihood7, $LogLikelihood8, $LogLikelihood9, 
$LogLikelihood10, $LogLikelihood11, $LogLikelihood12, $LogLikelihood13, 
$LogLikelihood14 where each of the numbers cooresponds to the above
described models.

=item 7

The function errorCode() returns 0 if the last operation was successful.
It will return an integer starting with 1 if the last operation failed. 
This indicates that statistic.pl should abort. It will return an integer 
starting with 2 to indicate a warning should be issued. This does not 
cause cause statistic.pl to abort. However, a warning after 
calculateStatistic() will cause the 4gram which generated that warning to 
be ignored by statistic.pl.

=item 8

The function errorString() returns the text of an error message. 

=back

=head2 Writing your own statistic model module

Any measure module being used by statistic.pl must follow this  
convention. In order to make it easier to build 4d measures, we provide 
a 4d specific functions that can be embedded within the measure module in 
order to carry out common operations in calculating the values of such 
measures. 

=over 4

=item 1 

The filename should have an extension of .pm. Usually the name of 
the file should be Statistic.pm, where "Statistic" is the name of the
particular statistic you are writing.

=item 2 

Let us say you have named your file Statistic.pm. The first line of the
file should declare that its a package of the same name as the filename.
Thus the first line of the file Statistic.pm should be...

   package Statistic;

=item 3

To use the measure4d.pm module you need to include it in your 
Statistic.pm module. 

A small code snippet to ensure that it is included is as follows:

 #  Check to see if the statistic.pm module can see the
 #  ll4modeld.pm module ... if not see if it can be found.

 my $module = "ll4model.pm"; my $modulename = "ll4model.pm";
 if( !( -f $modulename ) ) {
    my $found = 0;
    #  Check each of the PATHS to see if the module is there
    foreach (split(/:/, $ENV{PATH})) {
 	$module = $_ . "/" . $modulename;
	if ( -f $module ) { $found = 1; last; }
    }
    # if still not found anywhere, quit!
    if ( ! $found ) { print "Could not find $modulename.\n"; exit; }
 }

 # IMPORTANT : now include the module into the current package    
 require $module;

=item 4 

You need to implement at least two functions in your package

   i)   initializeStatistic()
   ii)  calculateStatistic()


Function initializeStatistic() is passed the a set of parameters 
that include the 4gram size which is 4, the total number of 
4grams in the corpus, the number of frequency combinations 
and an array containing the frequency combinations. More detail 
of these parameters are described above in the description of 
the ll4model::initializeStatistic function.

These paramters can be passed directly into the ll4model.pm
module's function ll4model::intializeStatistic. For example: 

 sub initializeStatistic {
 
     ll4model::initializeStatistic(@_);
 
 }

where @_ contains the input parameters.

This function is called before any calls to the function
calculateStatistic() and can be used by the statistic library 
to set up any values that may be required for the calculations
later. This function is not expected to return anything. If an 
error occurs, it can be reported through the mechanisms described 
below. 

The other mandatory function is calculateStatistic(). This is
passed an array containing the frequency values for an ngram as
found in the input n-gram file. 

Function calculateStatistic() is expected to return a (possibly
floating) value as the value of the statistical measure calculated
using the frequency values passed to it. 

There exists two main functions in the module ll4model.pm in order
to help calculate the 4gram statistic. 

   1.  ll4model::checkObservedValues(@frequencies)
   2.  ll4model::checkExpectedValues();

The function ll4model::checkObservedValues will return a 1.
If it does not then there existed an error in the calculation 
of these values and zero should be returned. An example of how 
this can be used is as follows:
  
if( !( $check  = ll4model::checkObservedValues(@_) ) ) {
    return(0);
}
 

where @_ is the parameters sent to calculateStatistic from statistic.pl.
A more detailed description of this function can be seen above.

The function ll4model::checkExpectedValues will return a 1.
If it does not then there existed an error in the calculation of 
these values and zero should be returned. An example of how this
can be used is as follows:

if( !( $check = ll4model::checkExpectedValues() ) ) {
	return(0);
}

When a library is loaded, statistic.pl checks for initializeStatistc
and calcualteStatistic functions: if they are not implemented, then 
an error is reported and the program quits. 

The function getLogLikelihoodValues will return a list of the 
log likelihood scores for each of the assumption models. An
example of how htis can be used is as follows:

if( ! ( ($LogLikelihood1,  $LogLikelihood2,  $LogLikelihood3,  
    $LogLikelihood4, $LogLikelihood5,  $LogLikelihood6,  $LogLikelihood7,  
    $LogLikelihood8, $LogLikelihood9,  $LogLikelihood10, $LogLikelihood11, 
    $LogLikelihood12, $LogLikelihood13, $LogLikelihood14) 
    = ll4model::getLogLikelihoodValues())) {
    
    return(0);
}

=item 5 

Program statistic.pl also supports three other functions that are
not mandatory, but may be implemented by the user. These are:

     i) errorCode()
    ii) errorString()
   iii) getStatisticName()

Function errorCode, if implemented, is called immediately after 
the call to function initializeStatistic() and immediately after 
every call to function calculateStatistic(). 

The ll4model.pm module implements both ll4model::errorCode() 
and ll4model::errorString().

The errorCode() and errorString() methods that are implmented in
your Statistic.pm modoule can return the value returned by the 
ll4model::errorCode() and ll4model::errorString() functions.

An example of this is below:

sub errorCode
{ 
  return ll4model::errorCode();
}

sub errorString
{
    return ll4model::errorString();
}

The third function that may be implemented is getStatisticName(). 
If this function is implemented, it is expected to return a string 
containing the name of the statistic being implmented. This string 
is used in the formatted output of statistic.pl. If this function 
is not implemented, then the statistic file name entered on the 
commandline is used in the formatted output. 

Note that all three functions described in this section are first
checked for existence before being called. So, if the user elects
to not implement these functions, no harm will be done. However, we
strongly recommend the implementation of at least the function
errorCode() since this is the only way for the statistic library to
report errors to the user. 

=item 6 

Having implemented the two mandatory functions (in point 4 above)
and zero or more of the three non-mandatory functions (in point 4
above), one must make these functions available outside the
package. To do so, one has to export them, thusly. 

For this, first include the Exporter package by including the
following line in the program

   require Exporter;

Now include the following line to inherit Exporter's functions:

   @ISA = qw ( Exporter );

Now export the various functions implemented so that they are
accessible outside this package, by adding the following line
(assume that you have implemented only the two mandatory
functions): 

   @EXPORT = qw( initializeStatistic calculateStatistic );

If you implement say the errorCode() and errorString() functions
too, you may export them like so: 

   @EXPORT = qw( initializeStatistic calculateStatistic errorCode errorString );
   
Note that the user may implement other functions too, and may
export them if he so wishes, but since statistic.pl is not
expecting anything besides the five functions above, doing so would
have no effect on statistic.pl. 

=item 7 

Finally, at the end of everything, add the line

   1;

   This will ensure that the LAST line of the file returns a 
   true value, and is necessary so that when this package is 
   loaded, it returns a TRUE value.

=back

=head2 Errors to look out for:

=over 4

=item 1 

The filename does not end with a .pm.

=item 2 

The rest of the filename (besides the extension) does not match the package
name (declared in the first line of the file). Remember its case sensitive!

=item 3 

The five functions (2 mandatory, 3 non-mandatory) must have their
names match EXACTLY with those shown above. Again, names are all
case sensitive.

=item 4 

The last line of the file is not "1;". This is necessary, and easily
overlooked!

=back

=head1 AUTHORS

    Ted Pedersen (tpederse@umn.edu)
    Satanjeev Banerjee <banerjee@cs.cmu.edu>
    Bridget McInnes (bthomson@d.umn.edu)
    
=head1 BUGS

=head1 SEE ALSO

 home page:    http://www.d.umn.edu/~tpederse/nsp.html

 mailing list: http://groups.yahoo.com/group/ngram/

=head1 COPYRIGHT

Copyright (C) 2004 Ted Pedersen and Bridget McInnes

Permission is granted to copy, distribute and/or modify this  document  
under the terms of the GNU Free Documentation License, Version 1.2 or  any  
later version published by the Free Software Foundation; with no  
Invariant Sections, no Front-Cover Texts, and no Back-Cover Texts.

Note: a copy of the GNU Free Documentation License is available on the web   
at L<http://www.gnu.org/copyleft/fdl.html> and is included in this    
distribution as FDL.txt. 

=cut

package ll4model;
require Exporter;

@ISA = qw ( Exporter );
@EXPORT = qw( initializeStatistic checkObservedValues checkExpectedValues getLogLikelihodValues getMarginalTotals getTotal4grams errorCode errorString );

my $n1pp = 0; my $np1p = 0; my $npp1 = 0; my $n11p = 0; my $np11 = 0;
my $n1p1 = 0; my $n2pp = 0; my $np2p = 0; my $npp2 = 0; my $nppp = 0;

# function to set up various variables before the actual computation
# starts. also to check if we are being given 4grams, and if our
# frequency combinations are enough to do the computation

sub initializeStatistic
{
    ($ngram, $npppp, $combIndex, @freqComb) = @_;
    
    $errorCodeNumber = 0;
    $errorMessage = "";
    
    # check if ngram !=4 . statistic only defined for ngram = 4. 
    if ($ngram != 4) {
	$errorMessage = "This statistic is only available for 4grams!";
	$errorCodeNumber = 1;  return;
    }
    
     # The total number of 4grams should not be less than equal to 0
    if ($npppp <= 0) { 
	$errorMessage = "Total number of 4grams ($npppp) must be greater than 0.";
	$errorCodeNumber = 10;  return;
    }

    # figure out from the @freqComb array if the frequency values we
    # are going to get are indeed the ones we need. 
    # while we figure this out, we shall also note which of the indices of the 
    # array passed to function calculateStatistic are the ones we want.

    for (my $i = 0; $i < $combIndex; $i++)
    {
	$str="";
        foreach(@{$freqComb[$i]}[1..$freqComb[$i][0]]) { $str.=$_."#"; }
        if($str eq "0#1#2#3#") {  $n1111Index=$i; }
        if($str eq "0#")       {  $n1pppIndex=$i; }
        if($str eq "1#")       {  $np1ppIndex=$i; }
	if($str eq "2#")       {  $npp1pIndex=$i; }
	if($str eq "3#")       {  $nppp1Index=$i; }
        if($str eq "0#1#")     {  $n11ppIndex=$i; }
	if($str eq "0#2#")     {  $n1p1pIndex=$i; }
	if($str eq "0#3#")     {  $n1pp1Index=$i; }
        if($str eq "1#2#")     {  $np11pIndex=$i; }
        if($str eq "1#3#")     {  $np1p1Index=$i; }
        if($str eq "2#3#")     {  $npp11Index=$i; }
        if($str eq "0#1#2#")   {  $n111pIndex=$i; }
        if($str eq "0#1#3#")   {  $n11p1Index=$i; }
        if($str eq "0#2#3#")   {  $n1p11Index=$i; }
        if($str eq "1#2#3#")   {  $np111Index=$i; }
    }
    
    if (!(defined $n1111Index)) { $errorCodeNumber = 100; $errorMessage = "Frequency combination \"0 1 2 3\" missing!\n"; }
    if (!(defined $n1pppIndex)) { $errorCodeNumber = 101; $errorMessage = "Frequency combination \"0\" missing!\n";   }
    if (!(defined $np1ppIndex)) { $errorCodeNumber = 102; $errorMessage = "Frequency combination \"1\" missing!\n";   }
    if (!(defined $npp1pIndex)) { $errorCodeNumber = 103; $errorMessage = "Frequency combination \"2\" missing!\n";   }
    if (!(defined $nppp1Index)) { $errorCodeNumber = 104; $errorMessage = "Frequency combination \"3\" missing!\n";   }
    if (!(defined $n11ppIndex)) { $errorCodeNumber = 105; $errorMessage = "Frequency combination \"0 1\" missing!\n"; }
    if (!(defined $n1p1pIndex)) { $errorCodeNumber = 106; $errorMessage = "Frequency combination \"0 2\" missing!\n"; }
    if (!(defined $n1pp1Index)) { $errorCodeNumber = 107; $errorMessage = "Frequency combination \"0 3\" missing!\n"; }
    if (!(defined $np11pIndex)) { $errorCodeNumber = 108; $errorMessage = "Frequency combination \"1 2\" missing!\n"; }
    if (!(defined $np1p1Index)) { $errorCodeNumber = 109; $errorMessage = "Frequency combination \"1 3\" missing!\n"; }
    if (!(defined $npp11Index)) { $errorCodeNumber = 110; $errorMessage = "Frequency combination \"2 3\" missing!\n"; }
    if (!(defined $n111pIndex)) { $errorCodeNumber = 111; $errorMessage = "Frequency combination \"0 1 2\" missing!\n"; }
    if (!(defined $n11p1Index)) { $errorCodeNumber = 112; $errorMessage = "Frequency combination \"0 1 3\" missing!\n"; }
    if (!(defined $n1p11Index)) { $errorCodeNumber = 113; $errorMessage = "Frequency combination \"0 2 3\" missing!\n"; }
    if (!(defined $np111Index)) { $errorCodeNumber = 114; $errorMessage = "Frequency combination \"1 2 3\" missing!\n"; }

}

# function to ensure the marginal values are reasonable with
# respect to each other and if so calculate the observed values
sub checkObservedValues() 
{
    $n1111=$_[$n1111Index];
    $n1ppp=$_[$n1pppIndex];
    $np1pp=$_[$np1ppIndex];
    $npp1p=$_[$npp1pIndex];
    $nppp1=$_[$nppp1Index];
    $n11pp=$_[$n11ppIndex];
    $n1p1p=$_[$n1p1pIndex];
    $n1pp1=$_[$n1pp1Index];
    $np11p=$_[$np11pIndex];
    $np1p1=$_[$np1p1Index];
    $npp11=$_[$npp11Index];
    $n111p=$_[$n111pIndex];
    $n11p1=$_[$n11p1Index];
    $n1p11=$_[$n1p11Index];
    $np111=$_[$np111Index];
    $n1112=$n111p-$n1111;
    $n1121=$n11p1-$n1111;
    
    $n1122=$n11pp-$n1111-$n1121-$n1112;
    $n1211=$n1p11-$n1111;
    $n1212=$n1p1p-$n1111-$n1112-$n1211;
    $n1221=$n1pp1-$n1111-$n1211-$n1121;
    $n1222=$n1ppp-$n1111-$n1211-$n1121-$n1112;
    $n2111=$np111-$n1111; 
    $n2112=$np11p-$n1111-$n2111-$n1112;
    $n2121=$np1p1-$n1111-$n2111-$n1121;
    $n2122=$np1pp-$n1111-$n2111-$n1121-$n1112;
    $n2211=$npp11-$n1111-$n2111-$n1211;
    $n2212=$npp1p-$n1111-$n2111-$n1211-$n1112;
    $n2221=$nppp1-$n1111-$n2111-$n1211-$n1121;
    $n2222=$npppp-($n1111+$n1112+$n1121+$n1122+$n1211+$n1212+$n1221+$n1222+$n2111+$n2112+$n2121+$n2122+$n2211+$n2212+$n2221);
    
    $n2ppp=$npppp-$n1ppp;
    $np2pp=$npppp-$np1pp;
    $npp2p=$npppp-$npp1p;
    $nppp2=$npppp-$nppp1;

    # n1111 should be greater than equal to zero 
    if ($n1111 <= 0)    {
	$errorMessage = "Frequency value ($n1111) must not be negative.";
	$errorCodeNumber = 200;  return;
    }
    
    # n1111 frequency should be less than or equal to total4grams
    if ($n1111> $npppp) {
	$errorMessage = "Frequency value ($n1111) must not exceed total number of 4grams.";
	$errorCodeNumber = 201;  return;
    }
    
    # joint frequency n1111 should be less than or equal to the marginal totals
    if ($n1111 > $n1ppp || $n1111 > $np1pp || $n1111 > $npp1p || $n1111 > $nppp1) {
	$errorMessage = "Frequency value of ngram ($n1111) must not exceed the marginal totals.";
	$errorCodeNumber = 202;  return;
    }
    
    # n1pp should be greater than or equal to zero 
    if ($n1ppp <= 0) {
	$errorMessage = "Marginal total value ($n1ppp) must not be negative.";
	$errorCodeNumber = 203;  return;
    }

    # n1ppp should be less than or equal to total4grams
    if ($n1ppp > $npppp) {
	$errorMessage = "Marginal total value ($n1ppp) must not exceed total number of 4grams ($npppp).";
	$errorCodeNumber = 204;  return;
    }

    # np1pp should be greater than or equal to zero
    if ($np1pp <= 0) {
	$errorMessage = "Marginal total value ($np1pp) must not be negative.";
        $errorCodeNumber = 205;  return;
    }

    # np1pp should be less than or equal to total4grams
    if ($np1pp > $npppp) {
	$errorMessage = "Marginal total value ($np1pp) must not exceed total number of 4grams.";
        $errorCodeNumber = 206;  return;
    }
    
    # npp1p should be greater than or equal to zero 
    if ($npp1p <= 0) {
	$errorMessage = "Marginal total value ($npp1p) must not be negative.";
	$errorCodeNumber = 207;  return;
    }
    
    # npp1p should be less than or equal to total4grams
    if ($npp1p > $npppp) {
	$errorMessage = "Marginal total value ($npp1p) must not exceed total number of 4grams.";
	$errorCodeNumber = 208;  return;
    }

    # nppp1p should be greater than or equal to zero 
    if ($nppp1 <= 0) {
	$errorMessage = "Marginal total value ($nppp1) must not be negative.";
	$errorCodeNumber = 209;  return;
    }
    
    # nppp1p should be less than or equal to total4grams
    if ($nppp1 > $npppp) {
	$errorMessage = "Marginal total value ($nppp1) must not exceed total number of 4grams.";
	$errorCodeNumber = 210;  return;
    }

    #  Everything looks good so we can return the observed values
    @values = ($n1111, $n1112, $n1121, $n1122, $n1211, $n1212, $n1221, $n1222, 
	       $n2111, $n2112, $n2121, $n2122, $n2211, $n2212, $n2221, $n2222);
    return 1;
}

#  Calculates, checks and returns the expected values 
#  and calculates the loglikelihood values as it goes 
sub checkExpectedValues
{
    #  initialize the log likelihood scores
    $LogLikelihood1 = 0; $LogLikelihood2 = 0; $LogLikelihood3 = 0; $LogLikelihood4 = 0;
    $LogLikelihood5 = 0; $LogLikelihood6 = 0; $LogLikelihood7 = 0; $LogLikelihood8 = 0;
    $LogLikelihood9 = 0; $LogLikelihood10= 0; $LogLikelihood11= 0; $LogLikelihood12= 0;
    $LogLikelihood13= 0; $LogLikelihood14= 0;
    
    #  calculate the log likelhood for "0 123" check
    $np111 = $n1111 + $n2111;    $np112 = $n1112 + $n2112;
    $np121 = $n1121 + $n2121;    $np122 = $n1122 + $n2122;
    $np211 = $n1211 + $n2211;    $np212 = $n1212 + $n2212;
    $np221 = $n1221 + $n2221;    $np222 = $n1222 + $n2222;

    $m1111=$n1ppp*$np111/($npppp);    $m1112=$n1ppp*$np112/($npppp);
    $m1121=$n1ppp*$np121/($npppp);    $m1122=$n1ppp*$np122/($npppp);
    $m1211=$n1ppp*$np211/($npppp);    $m1212=$n1ppp*$np212/($npppp);
    $m1221=$n1ppp*$np221/($npppp);    $m1222=$n1ppp*$np222/($npppp);
    $m2111=$n2ppp*$np111/($npppp);    $m2112=$n2ppp*$np112/($npppp);
    $m2121=$n2ppp*$np121/($npppp);    $m2122=$n2ppp*$np122/($npppp);
    $m2211=$n2ppp*$np211/($npppp);    $m2212=$n2ppp*$np212/($npppp);
    $m2221=$n2ppp*$np221/($npppp);    $m2222=$n2ppp*$np222/($npppp);
    
    #  check the marginals
    $check = CheckMarginals();
    if($check == 1) { return; }
    
    #  get the log likelihood value
    $LogLikelihood1 = getLogLikelihood();

    #  calculate the log likelhood value for "01 2 3" check
    $n12pp = $n1211 + $n1212 + $n1221 + $n1222;
    $n21pp = $n2111 + $n2112 + $n2121 + $n2122;
    $n22pp = $n2211 + $n2212 + $n2221 + $n2222;
    
    $m1111=$n11pp*$npp1p*$nppp1/($npppp**2);    $m1112=$n11pp*$npp1p*$nppp2/($npppp**2);
    $m1121=$n11pp*$npp2p*$nppp1/($npppp**2);    $m1122=$n11pp*$npp2p*$nppp2/($npppp**2);
    $m1211=$n12pp*$npp1p*$nppp1/($npppp**2);    $m1212=$n12pp*$npp1p*$nppp2/($npppp**2);
    $m1221=$n12pp*$npp2p*$nppp1/($npppp**2);    $m1222=$n12pp*$npp2p*$nppp2/($npppp**2);
    $m2111=$n21pp*$npp1p*$nppp1/($npppp**2);    $m2112=$n21pp*$npp1p*$nppp2/($npppp**2);
    $m2121=$n21pp*$npp2p*$nppp1/($npppp**2);    $m2122=$n21pp*$npp2p*$nppp2/($npppp**2);
    $m2211=$n22pp*$npp1p*$nppp1/($npppp**2);    $m2212=$n22pp*$npp1p*$nppp2/($npppp**2);
    $m2221=$n22pp*$npp2p*$nppp1/($npppp**2);    $m2222=$n22pp*$npp2p*$nppp2/($npppp**2);
    
    #  check the marginals
    $check = CheckMarginals();
    if($check == 1) { return; }
    
    #  get the log likelihood value
    $LogLikelihood2 = getLogLikelihood();
    
    #  calculate the loglikelihood value for : "0 1 23" check

    $npp12 = $n1112 + $n1212 + $n2112 + $n2212;
    $npp21 = $n1121 + $n1221 + $n2121 + $n2221;
    $npp22 = $n1122 + $n1222 + $n2122 + $n2222;
    
    $m1111=$n1ppp*$np1pp*$npp11/($npppp**2);    $m1112=$n1ppp*$np1pp*$npp12/($npppp**2);  
    $m1121=$n1ppp*$np1pp*$npp21/($npppp**2);    $m1122=$n1ppp*$np1pp*$npp22/($npppp**2);
    $m1211=$n1ppp*$np2pp*$npp11/($npppp**2);    $m1212=$n1ppp*$np2pp*$npp12/($npppp**2);  
    $m1221=$n1ppp*$np2pp*$npp21/($npppp**2);    $m1222=$n1ppp*$np2pp*$npp22/($npppp**2);
    $m2111=$n2ppp*$np1pp*$npp11/($npppp**2);    $m2112=$n2ppp*$np1pp*$npp12/($npppp**2);  
    $m2121=$n2ppp*$np1pp*$npp21/($npppp**2);    $m2122=$n2ppp*$np1pp*$npp22/($npppp**2);
    $m2211=$n2ppp*$np2pp*$npp11/($npppp**2);    $m2212=$n2ppp*$np2pp*$npp12/($npppp**2);  
    $m2221=$n2ppp*$np2pp*$npp21/($npppp**2);    $m2222=$n2ppp*$np2pp*$npp22/($npppp**2);
    
    #  check the marginals
    $check = CheckMarginals();
    if($check == 1) { return; }
    
    #  get the log likelihood value
    $LogLikelihood3 = getLogLikelihood();
        
    # calculate the loglikelihood value for : "0 12 3" check

    $np12p = $n1121 + $n1122 + $n2121 + $n2122;
    $np21p = $n1211 + $n1212 + $n2211 + $n2212;
    $np22p = $n1221 + $n1222 + $n2221 + $n2222;
    
    $m1111=$n1ppp*$np11p*$nppp1/($npppp**2);    $m1112=$n1ppp*$np11p*$nppp2/($npppp**2);
    $m1121=$n1ppp*$np12p*$nppp1/($npppp**2);    $m1122=$n1ppp*$np12p*$nppp2/($npppp**2);  
    $m1211=$n1ppp*$np21p*$nppp1/($npppp**2);    $m1212=$n1ppp*$np21p*$nppp2/($npppp**2);  
    $m1221=$n1ppp*$np22p*$nppp1/($npppp**2);    $m1222=$n1ppp*$np22p*$nppp2/($npppp**2);
    $m2111=$n2ppp*$np11p*$nppp1/($npppp**2);    $m2112=$n2ppp*$np11p*$nppp2/($npppp**2);
    $m2121=$n2ppp*$np12p*$nppp1/($npppp**2);    $m2122=$n2ppp*$np12p*$nppp2/($npppp**2);  
    $m2211=$n2ppp*$np21p*$nppp1/($npppp**2);    $m2212=$n2ppp*$np21p*$nppp2/($npppp**2);  
    $m2221=$n2ppp*$np22p*$nppp1/($npppp**2);    $m2222=$n2ppp*$np22p*$nppp2/($npppp**2);

    #  check the marginals
    $check = CheckMarginals();
    if($check == 1) { return; }
    
    #  get the log likelihood value
    $LogLikelihood4 = getLogLikelihood();
        
    #  calculate the loglikelihood value for : "012 3" check

    $n112p = $n1121 + $n1122;    $n121p = $n1211 + $n1212;
    $n122p = $n1221 + $n1222;    $n211p = $n2111 + $n2112;
    $n212p = $n2121 + $n2122;    $n221p = $n2211 + $n2212;
    $n222p = $n2221 + $n2222;
    
    $m1111=$n111p*$nppp1/($npppp);    $m1112=$n111p*$nppp2/($npppp);
    $m1121=$n112p*$nppp1/($npppp);    $m1122=$n112p*$nppp2/($npppp);
    $m1211=$n121p*$nppp1/($npppp);    $m1212=$n121p*$nppp2/($npppp);
    $m1221=$n122p*$nppp1/($npppp);    $m1222=$n122p*$nppp2/($npppp);
    $m2111=$n211p*$nppp1/($npppp);    $m2112=$n211p*$nppp2/($npppp);
    $m2121=$n212p*$nppp1/($npppp);    $m2122=$n212p*$nppp2/($npppp);
    $m2211=$n221p*$nppp1/($npppp);    $m2212=$n221p*$nppp2/($npppp);
    $m2221=$n222p*$nppp1/($npppp);    $m2222=$n222p*$nppp2/($npppp);

    #  check the marginals
    $check = CheckMarginals();
    if($check == 1) { return; }
    
    #  get the log likelihood value
    $LogLikelihood5 = getLogLikelihood();
    
    #  calculate the loglikelihood values for : "0 13 2" check
    
    $np1p2 = $n1112 + $n1122 + $n2112 + $n2122;
    $np2p1 = $n1211 + $n1221 + $n2211 + $n2221;
    $np2p2 = $n1212 + $n1222 + $n2212 + $n2222;
    
    $m1111=$n1ppp*$np1p1*$npp1p/($npppp**2);    $m1112=$n1ppp*$np1p2*$npp1p/($npppp**2);
    $m1121=$n1ppp*$np1p1*$npp2p/($npppp**2);    $m1122=$n1ppp*$np1p2*$npp2p/($npppp**2);
    $m1211=$n1ppp*$np2p1*$npp1p/($npppp**2);    $m1212=$n1ppp*$np2p2*$npp1p/($npppp**2);
    $m1221=$n1ppp*$np2p1*$npp2p/($npppp**2);    $m1222=$n1ppp*$np2p2*$npp2p/($npppp**2);
    $m2111=$n2ppp*$np1p1*$npp1p/($npppp**2);    $m2112=$n2ppp*$np1p2*$npp1p/($npppp**2);
    $m2121=$n2ppp*$np1p1*$npp2p/($npppp**2);    $m2122=$n2ppp*$np1p2*$npp2p/($npppp**2);
    $m2211=$n2ppp*$np2p1*$npp1p/($npppp**2);    $m2212=$n2ppp*$np2p2*$npp1p/($npppp**2);
    $m2221=$n2ppp*$np2p1*$npp2p/($npppp**2);    $m2222=$n2ppp*$np2p2*$npp2p/($npppp**2);

    #  check the marginals
    $check = CheckMarginals();
    if($check == 1) { return; }
    
    #  get the log likelihood value
    $LogLikelihood6 = getLogLikelihood();
        
    #  calculate the loglikelihood values for : "02 13" check

    $n1p2p = $n1121 + $n1122 + $n1221 + $n1222;
    $n2p1p = $n2111 + $n2112 + $n2211 + $n2212;
    $n2p2p = $n2121 + $n2122 + $n2221 + $n2222;
    
    $np1p2 = $n1112 + $n1122 + $n2112 + $n2122;
    $np2p1 = $n1211 + $n1221 + $n2211 + $n2221;
    $np2p2 = $n1212 + $n1222 + $n2212 + $n2222;
    
    
    $m1111=$n1p1p*$np1p1/($npppp);    $m1112=$n1p1p*$np1p2/($npppp);
    $m1121=$n1p2p*$np1p1/($npppp);    $m1122=$n1p2p*$np1p2/($npppp);
    $m1211=$n1p1p*$np2p1/($npppp);    $m1212=$n1p1p*$np2p2/($npppp);
    $m1221=$n1p2p*$np2p1/($npppp);    $m1222=$n1p2p*$np2p2/($npppp);
    $m2111=$n2p1p*$np1p1/($npppp);    $m2112=$n2p1p*$np1p2/($npppp);
    $m2121=$n2p2p*$np1p1/($npppp);    $m2122=$n2p2p*$np1p2/($npppp);
    $m2211=$n2p1p*$np2p1/($npppp);    $m2212=$n2p1p*$np2p2/($npppp);
    $m2221=$n2p2p*$np2p1/($npppp);    $m2222=$n2p2p*$np2p2/($npppp);

    #  check the marginals
    $check = CheckMarginals();
    if($check == 1) { return; }
    
    #  get the log likelihood value
    $LogLikelihood7 = getLogLikelihood();
    
    #  calculate the loglikelihood values for : "03 12" check
    
    $n1pp2 = $n1112 + $n1122 + $n1212 + $n1222;
    $n2pp1 = $n2111 + $n2121 + $n2211 + $n2221;
    $n2pp2 = $n2112 + $n2122 + $n2212 + $n2222;
    
    $np12p = $n1121 + $n1122 + $n2121 + $n2122;
    $np21p = $n1211 + $n1212 + $n2211 + $n2212;
    $np22p = $n1221 + $n1222 + $n2221 + $n2222;
    
    $m1111=$n1pp1*$np11p/($npppp);    $m1112=$n1pp2*$np11p/($npppp);
    $m1121=$n1pp1*$np12p/($npppp);    $m1122=$n1pp2*$np12p/($npppp);
    $m1211=$n1pp1*$np21p/($npppp);    $m1212=$n1pp2*$np21p/($npppp);
    $m1221=$n1pp1*$np22p/($npppp);    $m1222=$n1pp2*$np22p/($npppp);
    $m2111=$n2pp1*$np11p/($npppp);    $m2112=$n2pp2*$np11p/($npppp);
    $m2121=$n2pp1*$np12p/($npppp);    $m2122=$n2pp2*$np12p/($npppp);
    $m2211=$n2pp1*$np21p/($npppp);    $m2212=$n2pp2*$np21p/($npppp);
    $m2221=$n2pp1*$np22p/($npppp);    $m2222=$n2pp2*$np22p/($npppp);

    #  check the marginals
    $check = CheckMarginals();
    if($check == 1) { return; }
    
    #  get the log likelihood value
    $LogLikelihood8 = getLogLikelihood();
    
    #  calculate the loglikelihood values for : "023 1" check
   
    $n1p12 = $n1112 + $n1212;    $n1p21 = $n1121 + $n1221;
    $n1p22 = $n1122 + $n1222;    $n2p11 = $n2111 + $n2211;
    $n2p12 = $n2112 + $n2212;    $n2p21 = $n2121 + $n2221;
    $n2p22 = $n2122 + $n2222;
    
    $m1111=$n1p11*$np1pp/($npppp);    $m1112=$n1p12*$np1pp/($npppp);
    $m1121=$n1p21*$np1pp/($npppp);    $m1122=$n1p22*$np1pp/($npppp);
    $m1211=$n1p11*$np2pp/($npppp);    $m1212=$n1p12*$np2pp/($npppp);
    $m1221=$n1p21*$np2pp/($npppp);    $m1222=$n1p22*$np2pp/($npppp);
    $m2111=$n2p11*$np1pp/($npppp);    $m2112=$n2p12*$np1pp/($npppp);
    $m2121=$n2p21*$np1pp/($npppp);    $m2122=$n2p22*$np1pp/($npppp);
    $m2211=$n2p11*$np2pp/($npppp);    $m2212=$n2p12*$np2pp/($npppp);
    $m2221=$n2p21*$np2pp/($npppp);    $m2222=$n2p22*$np2pp/($npppp);

    #  check the marginals
    $check = CheckMarginals();
    if($check == 1) { return; }
    
    #  get the log likelihood value
    $LogLikelihood9 = getLogLikelihood();
    
    #  calculate the loglikelihood values for : "013 2" check
    
    $n11p2 = $n1112 + $n1122;    $n12p1 = $n1211 + $n1221;
    $n12p2 = $n1212 + $n1222;    $n21p1 = $n2111 + $n2121;
    $n21p2 = $n2112 + $n2122;    $n22p1 = $n2211 + $n2221;
    $n22p2 = $n2212 + $n2222;
    
    $m1111=$n11p1*$npp1p/($npppp);    $m1112=$n11p2*$npp1p/($npppp);
    $m1121=$n11p1*$npp2p/($npppp);    $m1122=$n11p2*$npp2p/($npppp);
    $m1211=$n12p1*$npp1p/($npppp);    $m1212=$n12p2*$npp1p/($npppp);
    $m1221=$n12p1*$npp2p/($npppp);    $m1222=$n12p2*$npp2p/($npppp);
    $m2111=$n21p1*$npp1p/($npppp);    $m2112=$n21p2*$npp1p/($npppp);
    $m2121=$n21p1*$npp2p/($npppp);    $m2122=$n21p2*$npp2p/($npppp);
    $m2211=$n22p1*$npp1p/($npppp);    $m2212=$n22p2*$npp1p/($npppp);
    $m2221=$n22p1*$npp2p/($npppp);    $m2222=$n22p2*$npp2p/($npppp);

    #  check the marginals
    $check = CheckMarginals();
    if($check == 1) { return; }
    
    #  get the log likelihood value
    $LogLikelihood10 = getLogLikelihood();

    #  calculate the loglikelihood values for : "01 23" check
    $n12pp = $n1211+ $n1212 + $n1221 + $n1222;
    $n21pp = $n2111+ $n2112 + $n2121 + $n2122;
    $n22pp = $n2211+ $n2212 + $n2221 + $n2222;
    
    $npp12 = $n1112 + $n1212 + $n2112 + $n2212;
    $npp21 = $n1121 + $n1221 + $n2121 + $n2221;
    $npp22 = $n1122 + $n1222 + $n2122 + $n2222;
    
    
    $m1111=$n11pp*$npp11/($npppp);    $m1112=$n11pp*$npp12/($npppp);
    $m1121=$n11pp*$npp21/($npppp);    $m1122=$n11pp*$npp22/($npppp);
    $m1211=$n12pp*$npp11/($npppp);    $m1212=$n12pp*$npp12/($npppp);
    $m1221=$n12pp*$npp21/($npppp);    $m1222=$n12pp*$npp22/($npppp);
    $m2111=$n21pp*$npp11/($npppp);    $m2112=$n21pp*$npp12/($npppp);
    $m2121=$n21pp*$npp21/($npppp);    $m2122=$n21pp*$npp22/($npppp);
    $m2211=$n22pp*$npp11/($npppp);    $m2212=$n22pp*$npp12/($npppp);
    $m2221=$n22pp*$npp21/($npppp);    $m2222=$n22pp*$npp22/($npppp);

    #  check the marginals
    $check = CheckMarginals();
    if($check == 1) { return; }
    
    #  get the log likelihood value
    $LogLikelihood11 = getLogLikelihood();

    #  calculate the loglikelihood values for : "03 1 2" check
    $n1pp2 = $n1112 + $n1122 + $n1212 + $n1222;
    $n2pp1 = $n2111 + $n2121 + $n2211 + $n2221;
    $n2pp2 = $n2112 + $n2122 + $n2212 + $n2222;
	
    $m1111=$n1pp1*$np1pp*$npp1p/($npppp**2);    $m1112=$n1pp2*$np1pp*$npp1p/($npppp**2);
    $m1121=$n1pp1*$np1pp*$npp2p/($npppp**2);    $m1122=$n1pp2*$np1pp*$npp2p/($npppp**2);
    $m1211=$n1pp1*$np2pp*$npp1p/($npppp**2);    $m1212=$n1pp2*$np2pp*$npp1p/($npppp**2);
    $m1221=$n1pp1*$np2pp*$npp2p/($npppp**2);    $m1222=$n1pp2*$np2pp*$npp2p/($npppp**2);
    $m2111=$n2pp1*$np1pp*$npp1p/($npppp**2);    $m2112=$n2pp2*$np1pp*$npp1p/($npppp**2);
    $m2121=$n2pp1*$np1pp*$npp2p/($npppp**2);    $m2122=$n2pp2*$np1pp*$npp2p/($npppp**2);
    $m2211=$n2pp1*$np2pp*$npp1p/($npppp**2);    $m2212=$n2pp2*$np2pp*$npp1p/($npppp**2);
    $m2221=$n2pp1*$np2pp*$npp2p/($npppp**2);    $m2222=$n2pp2*$np2pp*$npp2p/($npppp**2);

    #  check the marginals
    $check = CheckMarginals();
    if($check == 1) { return; }
    
    #  get the log likelihood value
    $LogLikelihood12 = getLogLikelihood();

    #  calculate the loglikelihood values for : "02 1 3" check
    
    $n1p2p = $n1121 + $n1122 + $n1221 + $n1222;
    $n2p1p = $n2111 + $n2112 + $n2211 + $n2212;
    $n2p2p = $n2121 + $n2122 + $n2221 + $n2222;
    
    
    $m1111=$n1p1p*$np1pp*$nppp1/($npppp**2);    $m1112=$n1p1p*$np1pp*$nppp2/($npppp**2);
    $m1121=$n1p2p*$np1pp*$nppp1/($npppp**2);    $m1122=$n1p2p*$np1pp*$nppp2/($npppp**2);
    $m1211=$n1p1p*$np2pp*$nppp1/($npppp**2);    $m1212=$n1p1p*$np2pp*$nppp2/($npppp**2);
    $m1221=$n1p2p*$np2pp*$nppp1/($npppp**2);    $m1222=$n1p2p*$np2pp*$nppp2/($npppp**2);
    $m2111=$n2p1p*$np1pp*$nppp1/($npppp**2);    $m2112=$n2p1p*$np1pp*$nppp2/($npppp**2);
    $m2121=$n2p2p*$np1pp*$nppp1/($npppp**2);    $m2122=$n2p2p*$np1pp*$nppp2/($npppp**2);
    $m2211=$n2p1p*$np2pp*$nppp1/($npppp**2);    $m2212=$n2p1p*$np2pp*$nppp2/($npppp**2);
    $m2221=$n2p2p*$np2pp*$nppp1/($npppp**2);    $m2222=$n2p2p*$np2pp*$nppp2/($npppp**2);

    #  check the marginals
    $check = CheckMarginals();
    if($check == 1) { return; }
    
    #  get the log likelihood value
    $LogLikelihood13 = getLogLikelihood();
    
    #  calculate the log likelihood values for the model of independence check

    $m1111=$n1ppp*$np1pp*$npp1p*$nppp1/($npppp**3);    $m1112=$n1ppp*$np1pp*$npp1p*$nppp2/($npppp**3);
    $m1121=$n1ppp*$np1pp*$npp2p*$nppp1/($npppp**3);    $m1122=$n1ppp*$np1pp*$npp2p*$nppp2/($npppp**3);
    $m1211=$n1ppp*$np2pp*$npp1p*$nppp1/($npppp**3);    $m1212=$n1ppp*$np2pp*$npp1p*$nppp2/($npppp**3);
    $m1221=$n1ppp*$np2pp*$npp2p*$nppp1/($npppp**3);    $m1222=$n1ppp*$np2pp*$npp2p*$nppp2/($npppp**3);
    $m2111=$n2ppp*$np1pp*$npp1p*$nppp1/($npppp**3);    $m2112=$n2ppp*$np1pp*$npp1p*$nppp2/($npppp**3);
    $m2121=$n2ppp*$np1pp*$npp2p*$nppp1/($npppp**3);    $m2122=$n2ppp*$np1pp*$npp2p*$nppp2/($npppp**3);
    $m2211=$n2ppp*$np2pp*$npp1p*$nppp1/($npppp**3);    $m2212=$n2ppp*$np2pp*$npp1p*$nppp2/($npppp**3);
    $m2221=$n2ppp*$np2pp*$npp2p*$nppp1/($npppp**3);    $m2222=$n2ppp*$np2pp*$npp2p*$nppp2/($npppp**3);
    
    #  check the marginals
    $check = CheckMarginals();
    if($check == 1) { return; }
    
    #  get the log likelihood value
    $LogLikelihood14 = etLogLikelihood();
        
    return 1;
}

#  return the log likelihood values
sub getLogLikelihoodValues
{
    my @values = ($LogLikelihood1,  $LogLikelihood2,  $LogLikelihood3,  $LogLikelihood4, 
		  $LogLikelihood5,  $LogLikelihood6,  $LogLikelihood7,  $LogLikelihood8,  
		  $LogLikelihood9,  $LogLikelihood10, $LogLikelihood11, $LogLikelihood12, 
		  $LogLikelihood13, $LogLikelihood14);
    
    return @values;
}

#  calculate the log Likelihood score
sub getLogLikelihood
{
    my $logLikelihood = 0;

    # dont want ($n111 / $e111) to be 0 or less! flag error if so!
    if ( $n1111 ) { { $logLikelihood += $n1111 * log ( $n1111 / $m1111 ); }}
    if ( $n1112 ) { { $logLikelihood += $n1112 * log ( $n1112 / $m1112 ); }}
    if ( $n1121 ) { { $logLikelihood += $n1121 * log ( $n1121 / $m1121 ); }}
    if ( $n1122 ) { { $logLikelihood += $n1122 * log ( $n1122 / $m1122 ); }}
    if ( $n1211 ) { { $logLikelihood += $n1211 * log ( $n1211 / $m1211 ); }}
    if ( $n1212 ) { { $logLikelihood += $n1212 * log ( $n1212 / $m1212 ); }}
    if ( $n1221 ) { { $logLikelihood += $n1221 * log ( $n1221 / $m1221 ); }}
    if ( $n1222 ) { { $logLikelihood += $n1222 * log ( $n1222 / $m1222 ); }}
    if ( $n2111 ) { { $logLikelihood += $n2111 * log ( $n2111 / $m2111 ); }}
    if ( $n2112 ) { { $logLikelihood += $n2112 * log ( $n2112 / $m2112 ); }}
    if ( $n2121 ) { { $logLikelihood += $n2121 * log ( $n2121 / $m2121 ); }}
    if ( $n2122 ) { { $logLikelihood += $n2122 * log ( $n2122 / $m2122 ); }}
    if ( $n2211 ) { { $logLikelihood += $n2211 * log ( $n2211 / $m2211 ); }}
    if ( $n2212 ) { { $logLikelihood += $n2212 * log ( $n2212 / $m2212 ); }}
    if ( $n2221 ) { { $logLikelihood += $n2221 * log ( $n2221 / $m2221 ); }}
    if ( $n2222 ) { { $logLikelihood += $n2222 * log ( $n2222 / $m2222 ); }}
    
    return ($logLikelihood * 2);
}


sub CheckMarginals
{
    
    # dont want ($n1111 / $m1111) to be 0 or less! flag error if so!
    if ( $n1111 )  
    { 
	if ($m1111 == 0) {
	    $errorMessage = "Expected value in cell (1,1,1,1) must not be zero";
	    $errorCodeNumber = 231;  return 1;
	}

	if (($n1111 / $m1111)  <= 0) {
	    $errorMessage = "About to take log of negative value for cell (1,1,1,1)";
	    $errorCodeNumber = 232;  return 1;
	}
    }

    # dont want ($n1112 / $m1112) to be 0 or less! flag error if so!
    if ( $n1112 )  
    { 
	if ($m1112 == 0) {
	    $errorMessage = "Expected value in cell (1,1,1,2) must not be zero";
	    $errorCodeNumber = 233;  return 1;
	}

	if (($n1112 / $m1112)  <= 0) {
	    $errorMessage = "About to take log of negative value for cell (1,1,1,2)";
	    $errorCodeNumber = 234;  return 1;
	}
    }

    # dont want ($n1121 / $m1121) to be 0 or less! flag error if so!
    if ( $n1121 )  
    { 
	if ($m1121 == 0) {
	    $errorMessage = "Expected value in cell (1,1,2,1) must not be zero";
	    $errorCodeNumber = 235;  return 1;
	}

	if (($n1121 / $m1121)  <= 0) {
	    $errorMessage = "About to take log of negative value for cell (1,1,2,1)";
	    $errorCodeNumber = 236;  return 1;
	}
    }


    # dont want ($n1122 / $m1122) to be 0 or less! flag error if so!
    if ( $n1122 )  
    { 
	if ($m1122 == 0) {
	    $errorMessage = "Expected value in cell (1,1,2,2) must not be zero";
	    $errorCodeNumber = 237;  return 1;
	}

	if (($n1122 / $m1122)  <= 0) {
	    $errorMessage = "About to take log of negative value for cell (1,1,2,2)";
	    $errorCodeNumber = 238;  return 1;
	}
    }

    # dont want ($n1211 / $m1211) to be 0 or less! flag error if so!
    if ( $n1211 )  
    { 
	if ($m1211 == 0) {
	    $errorMessage = "Expected value in cell (1,2,1,1) must not be zero";
	    $errorCodeNumber = 239;  return 1;
	}

	if (($n1211 / $m1211)  <= 0) {
	    $errorMessage = "About to take log of negative value for cell (1,2,1,1)";
	    $errorCodeNumber = 240;  return 1;
	}
    }

    # dont want ($n1212 / $m1212) to be 0 or less! flag error if so!
    if ( $n1212 )  
    { 
	if ($m1212 == 0) {
	    $errorMessage = "Expected value in cell (1,2,1,2) must not be zero";
	    $errorCodeNumber = 241;  return 1;
	}

	if (($n1212 / $m1212)  <= 0) {
	    $errorMessage = "About to take log of negative value for cell (1,2,1,2)";
	    $errorCodeNumber = 242;  return 1;
	}
    }

    # dont want ($n1221 / $m1221) to be 0 or less! flag error if so!
    if ( $n1221 )  
    { 
	if ($m1221 == 0) {
	    $errorMessage = "Expected value in cell (1,2,2,1) must not be zero";
	    $errorCodeNumber = 243;  return 1;
	}

	if (($n1221 / $m1221)  <= 0) {
	    $errorMessage = "About to take log of negative value for cell (1,2,2,1)";
	    $errorCodeNumber = 244;  return 1;
	}
    }

    # dont want ($n1222 / $m1222) to be 0 or less! flag error if so!
    if ( $n1222 )  
    { 
	if ($m1222 == 0) {
	    $errorMessage = "Expected value in cell (1,2,2,2) must not be zero";
	    $errorCodeNumber = 245;  return 1;
	}

	if (($n1222 / $m1222)  <= 0) {
	    $errorMessage = "About to take log of negative value for cell (1,2,2,2)";
	    $errorCodeNumber = 246;  return 1;
	}
    }

    # dont want ($n2111 / $m2111) to be 0 or less! flag error if so!
    if ( $n2111 )  
    { 
	if ($m2111 == 0) {
	    $errorMessage = "Expected value in cell (2,1,1,1) must not be zero";
	    $errorCodeNumber = 247;  return 1;
	}

	if (($n2111 / $m2111)  <= 0) {
	    $errorMessage = "About to take log of negative value for cell (2,1,1,1)";
	    $errorCodeNumber = 248;  return 1;
	}
    }

    # dont want ($n2112 / $m2112) to be 0 or less! flag error if so!
    if ( $n2112 )  
    { 
	if ($m2112 == 0) {
	    $errorMessage = "Expected value in cell (2,1,1,2) must not be zero";
	    $errorCodeNumber = 249;  return 1;
	}

	if (($n2112 / $m2112)  <= 0) {
	    $errorMessage = "About to take log of negative value for cell (2,1,1,2)";
	    $errorCodeNumber = 250;  return 1;
	}
    }

    # dont want ($n2121 / $m2121) to be 0 or less! flag error if so!
    if ( $n2121 )  
    { 
	if ($m2121 == 0) {
	    $errorMessage = "Expected value in cell (2,1,2,1) must not be zero";
	    $errorCodeNumber = 251;  return 1;
	}

	if (($n2121 / $m2121)  <= 0) {
	    $errorMessage = "About to take log of negative value for cell (2,1,2,1)";
	    $errorCodeNumber = 252;  return 1;
	}
    }

    # dont want ($n2122 / $m2122) to be 0 or less! flag error if so!
    if ( $n2122 )  
    { 
	if ($m2122 == 0) {
	    $errorMessage = "Expected value in cell (2,1,2,2) must not be zero";
	    $errorCodeNumber = 253;  return 1;
	}

	if (($n2122 / $m2122)  <= 0) {
	    $errorMessage = "About to take log of negative value for cell (2,1,2,2)";
	    $errorCodeNumber = 254;  return 1;
	}
    }

    # dont want ($n2211 / $m2211) to be 0 or less! flag error if so!
    if ( $n2211 )  
    { 
	if ($m2211 == 0) {
	    $errorMessage = "Expected value in cell (2,2,1,1) must not be zero";
	    $errorCodeNumber = 255;  return 1;
	}

	if (($n2211 / $m2211)  <= 0) {
	    $errorMessage = "About to take log of negative value for cell (2,2,1,1)";
	    $errorCodeNumber = 256;  return 1;
	}
    }

    # dont want ($n1111 / $m1111) to be 0 or less! flag error if so!
    if ( $n2212 )  
    { 
	if ($m2212 == 0) {
	    $errorMessage = "Expected value in cell (2,2,1,2) must not be zero";
	    $errorCodeNumber = 257;  return 1;
	}

	if (($n2212 / $m2212)  <= 0) {
	    $errorMessage = "About to take log of negative value for cell (2,2,1,2)";
	    $errorCodeNumber = 258;  return 1;
	}
    }

    # dont want ($n2221 / $m2221) to be 0 or less! flag error if so!
    if ( $n2221 )  
    { 
	if ($m2221 == 0) {
	    $errorMessage = "Expected value in cell (2,2,2,1) must not be zero";
	    $errorCodeNumber = 259;  return 1;
	}

	if (($n2221 / $m2221)  <= 0) {
	    $errorMessage = "About to take log of negative value for cell (2,2,2,1)";
	    $errorCodeNumber = 260;  return 1;
	}
    }

    # dont want ($n2222 / $m2222) to be 0 or less! flag error if so!
    if ( $n2222 )  
    { 
	if ($m2222 == 0) {
	    $errorMessage = "Expected value in cell (2,2,2,2) must not be zero";
	    $errorCodeNumber = 261;  return 1;
	}

	if (($n2222 / $m2222)  <= 0) {
	    $errorMessage = "About to take log of negative value for cell (2,2,2,2)";
	    $errorCodeNumber = 262;  return 1;
	}
    }

    return 0;
}

#  Returns the marginal frequencies
sub getMarginalTotals {
    @values = ( $n1ppp, $np1pp, $npp1p, $nppp1, $n11pp, $n1p1p, $n1pp1, $np11p, $np1p1, $np11p, $n111p, $n11p1, $n1p11, $np111 );
    return @values;
}

#  Return the total number of 4grams
sub getTotal4grams {
    return $npppp;
}

# function to return the error code of the last operation and reset
# error code. useful if the error can be recovered from!
sub errorCode 
{ 
    my $temp = $errorCodeNumber;
    $errorCodeNumber = 0;
    return($temp); 
}


# function to return the error message of the last operation and reset
# the message string. useful if error can be recovered from!
sub errorString
{
    my $temp = $errorMessage;
    $errorMessage = "";
    return($temp);
}

1;
